<?php
/**
 * AmazonSearch MediaWiki extension  version 1.1.0
 *	for details please see: https://kimagurenote.net/kn/AmazonSearch
 *
 * Copyright (c) 2019-2020 Kimagurenote https://kimagurenote.net/
 * CodeSnippet.php sample code, Copyright (c) 2018 Amazon.com, Inc. or its affiliates.
 * Licensed under the Apache License, Version 2.0. https://opensource.org/licenses/Apache2.0
 * https://webservices.amazon.co.jp/paapi5/scratchpad/index.html
 *
 * Function:
 *	Mediawiki extension that make item links for several shopping site.
 *
 * Usage:
 *	<amazon>free keyword</amazon> on your wiki page.
 *	for details please see: https://kimagurenote.net/kn/AmazonSearch
 *
 * History:
 * 2020.03.02 Version 1.1.0
 *	Migrated PA-API 4.0 → 5.0
 *	License change: Revised BSD → Apache 2.0 because include Amazon sample code.
 * 2019.06.20 Version 1.0.1
 *	bypass mbstring if not defined.
 * 2019.06.20 Version 1.0.0
 *	1st release
 *
 * @file
 * @ingroup Extensions
 * @author Kimagurenote
 * @copyright © 2019 Kimagurenote
 * @license The BSD 3-Clause License
 */

if( !defined( 'MEDIAWIKI' ) ) {
	echo( "This file is an extension to the MediaWiki software and cannot be used standalone.\n" );
	die( 1 );
}

$wgExtensionCredits['parserhook'][] = array(
	'path'           => __FILE__,
	'name'           => 'AmazonSearch',
	'author'         => array(
		'[https://kimagurenote.net/ Kimagurenote]'
	),
	'url'            => 'https://kimagurenote.net/kn/AmazonSearch',
	'version'        => '1.1.0',
	'license-name'   => 'Apache 2.0',
	'descriptionmsg' => 'amazonsearch-desc'
);

/* Internationalization */
$wgMessagesDirs['AmazonSearch'] = __DIR__ . '/i18n';

/* for debug */
//define( 'AMAZONSEARCH_DEBUG', 'amazonsearch' );	/* for debug */
//$wgDebugLogGroups[AMAZONSEARCH_DEBUG] = $IP . '/extensions/AmazonSearch/debug.log';
//AMAZONSEARCH_DEBUG && error_reporting( -1 );
//AMAZONSEARCH_DEBUG && ini_set( 'display_errors', 1 );

/* default values */
define( 'AMAZONSEARCH_HOOKID', 'amazon' );			/* tag name */
define( 'AMAZONSEARCH_CLASS_DIV', 'amazonsearch' );	/* html div class name */

/* query type identifier */
define( 'AMAZONSEARCH_QUERYTYPE_ISBN', 'isbn' );
define( 'AMAZONSEARCH_QUERYTYPE_EAN', 'ean' );

/* operation parameter */
define( 'AMAZONSEARCH_NAMEMAXLEN', -1 );
define( 'AMAZONSEARCH_AUTHORMAXLEN', -1 );

/* HTML A tag target */
$wgAmazonSearchTarget = "_blank";

/* Product Advertising API URL list
 * https://images-na.ssl-images-amazon.com/images/G/09/associates/paapi/dg/index.html?CHAP_Intro_AAWS.html
 */
define( 'AMAZONSEARCH_APIURL_METHOD', 'POST' );
define( 'AMAZONSEARCH_APIURL_PROTOCOL', 'http://' );
define( 'AMAZONSEARCH_APIURL_PATH', '/paapi5/searchitems' );

/* for PA-API v5 */
$wgAmazonSearchAPITarget['SearchItems'] = "com.amazon.paapi5.v1.ProductAdvertisingAPIv1.SearchItems";

/* https://webservices.amazon.com/paapi5/documentation/common-request-parameters.html#host-and-region */
$wgAmazonSearchAPIHost = array(
	'ae' => 'webservices.amazon.ae',
	'au' => 'webservices.amazon.com.au',
	'br' => 'webservices.amazon.com.br',
	'ca' => 'webservices.amazon.ca',
	'de' => 'webservices.amazon.de',
	'es' => 'webservices.amazon.es',
	'fr' => 'webservices.amazon.fr',
	'in' => 'webservices.amazon.in',
	'it' => 'webservices.amazon.it',
	'jp' => 'webservices.amazon.co.jp',
	'mx' => 'webservices.amazon.com.mx',
	'sg' => 'webservices.amazon.sg',
	'tr' => 'webservices.amazon.com.tr',
	'uk' => 'webservices.amazon.co.uk',
	'us' => 'webservices.amazon.com'
);

$wgAmazonSearchAPIRegion = array(
	'ae' => 'eu-west-1',
	'au' => 'us-west-2',
	'br' => 'us-east-1',
	'ca' => 'us-east-1',
	'de' => 'eu-west-1',
	'es' => 'eu-west-1',
	'fr' => 'eu-west-1',
	'in' => 'eu-west-1',
	'it' => 'eu-west-1',
	'jp' => 'us-west-2',
	'mx' => 'us-east-1',
	'sg' => 'us-west-2',
	'tr' => 'eu-west-1',
	'uk' => 'eu-west-1',
	'us' => 'us-east-1'
);

/* https://webservices.amazon.com/paapi5/documentation/locale-reference.html */
$wgAmazonSearchAPIMarketplace = array(
	'ae' => 'www.amazon.ae',
	'au' => 'www.amazon.com.au',
	'br' => 'www.amazon.com.br',
	'ca' => 'www.amazon.ca',
	'de' => 'www.amazon.de',
	'es' => 'www.amazon.es',
	'fr' => 'www.amazon.fr',
	'in' => 'www.amazon.in',
	'it' => 'www.amazon.it',
	'jp' => 'www.amazon.co.jp',
	'mx' => 'www.amazon.com.mx',
	'sg' => 'www.amazon.sg',
	'tr' => 'www.amazon.com.tr',
	'uk' => 'www.amazon.co.uk',
	'us' => 'www.amazon.com'
);

# ae/au/br/ ca/de/es/fr/in/it/jp/mx/sg/tr/uk/us
$wgAmazonSearchDefaultRegion = 'jp';

/* tag classes */
$wgAmazonSearchClass['div'] = '';
$wgAmazonSearchClass['img'] = 'amazonsearchimg';
$wgAmazonSearchClass['name'] = 'amazonsearchname';
$wgAmazonSearchClass['author'] = 'amazonsearchauthor';
$wgAmazonSearchClass['buttons'] = 'amazonsearchbuttons';
$wgAmazonSearchClass['button'] = 'shoplinkamazon';

/* accepted parameters for pass to APIv5 */
/* https://webservices.amazon.com/paapi5/documentation/search-items.html */
$wgAmazonSearchAcceptArgs = array( 'Actor', 'Artist', 'Author', 'Brand', 'BrowseNodeId', 'Condition', 'DeliveryFlags', 'MaxPrice', 'Merchant', 'MinPrice', 'MinReviewsRating', 'MinSavingPercent', 'OfferCount', 'SearchIndex', 'SortBy', 'Title' );

/* MediaWiki set hook */
$wgHooks['ParserFirstCallInit'][] = 'wfAmazonSearchSetHook';

function wfAmazonSearchSetHook( $parser ) {
	$parser->setHook( AMAZONSEARCH_HOOKID, 'wfRenderAmazonSearch' );
	return true;
}

/* Main process - The callback function for embed shopping link
 *
 */
function wfRenderAmazonSearch( $value, $argv, $parser ) {
	global $wgAmazonSearchAPIHost, $wgAmazonSearchAPIRegion, $wgAmazonSearchAPIMarketplace, $wgAmazonSearchAPIKey, $wgAmazonSearchAPISecret, $wgAmazonSearchAssociateTag, $wgAmazonSearchDefaultRegion, $wgAmazonSearchDefaultMerchant, $wgAmazonSearchAcceptArgs, $wgAmazonSearchClass, $wgAmazonSearchMaxLen, $wgAmazonSearchTarget;

	/* for debug */	AMAZONSEARCH_DEBUG && ( $parser->disableCache() );

	/* search keyword */
	$value = wfAmazonSearchConvertEncoding(trim($value), 'UTF-8');

	/* require apiurl, appid and keyword(value) */
	if( empty($wgAmazonSearchAPIKey) || empty($wgAmazonSearchAPISecret) || empty($value) ) {
		/* for debug */	AMAZONSEARCH_DEBUG && ( wfDebugLog( AMAZONSEARCH_DEBUG, "wfRenderAmazonSearch(): some required parameter(s) empty.\n" ) );
		return '<!-- amazon>some required parameter(s) empty.</amazon -->';
	}

	/* region select */
	$region = (@$argv['region']) ? @$argv['region'] : $wgAmazonSearchDefaultRegion;
	if( empty($wgAmazonSearchAPIHost[$region]) ) {
		/* for debug */	AMAZONSEARCH_DEBUG && ( wfDebugLog( AMAZONSEARCH_DEBUG, "wfRenderAmazonSearch(): Region $region not defined.\n" ) );
		return '<!-- amazon>Region ' . htmlentities($region) . ' not defined.</amazon -->';
	}

	/* generate class for Amazon PA-API v5 */
	$paapi = new AmazonSearchPAAPI5SearchItems(
		$wgAmazonSearchAPIKey[$region] ? $wgAmazonSearchAPIKey[$region] : $wgAmazonSearchAPIKey,
		$wgAmazonSearchAPISecret[$region] ? $wgAmazonSearchAPISecret[$region] : $wgAmazonSearchAPISecret,
		$wgAmazonSearchAPIHost[$region],
		$wgAmazonSearchAPIRegion[$region]
	);

	/* generate API parameters */
	$atag = $apiparam['PartnerTag'] = isset($wgAmazonSearchAssociateTag[$region]) ? $wgAmazonSearchAssociateTag[$region] : $wgAmazonSearchAssociateTag;
	$apiparam['PartnerType'] = "Associates";
	$apiparam['Marketplace'] = $wgAmazonSearchAPIMarketplace[$region];
	$apiparam['SearchIndex'] = "All";
	/* if value seems ISBN */
	if( preg_match('/978\-[\d\-]+/', $value) ) {
		$value = str_replace('-', '', $value);
	}
	$apiparam['Keywords'] = $value;
	$apiparam['Resources'] = ["Images.Primary.Medium", "ItemInfo.Title", "ItemInfo.ByLineInfo"];
	$apiparam['Availability'] = 'Available';
	$apiparam['Condition'] = 'Any';
	$apiparam['Merchant'] = empty($wgAmazonSearchDefaultMerchant) ? 'All' : $wgAmazonSearchDefaultMerchant;

	/* override/more parameters if accepted (option) */
	if( !empty(@$argv) ) {
		foreach( $wgAmazonSearchAcceptArgs as $v ) {
			if( !empty(@$argv[strtolower($v)]) ) {
				$apiparam[$v] = @$argv[strtolower($v)];
				/* for debug */	AMAZONSEARCH_DEBUG && ( wfDebugLog( AMAZONSEARCH_DEBUG, "wfRenderAmazonSearch(): override parameter $v = " . $apiparam[$v] . "\n" ) );
			}
		}
	}

	/* generate params */
	$paapi->setPayload($apiparam);
	/* for debug */	AMAZONSEARCH_DEBUG && ( wfDebugLog( AMAZONSEARCH_DEBUG, "wfRenderAmazonSearch(): generated params as follows:\n" . var_export($apiparam, TRUE) . "\n\n" ) );
	$stream = $paapi->createStreamContext();

	/* call API */
	$apiurl = $paapi->getURL();
	if( ($ret = file_get_contents($apiurl, false, $stream)) === false ) {
		/* for debug */	AMAZONSEARCH_DEBUG && ( wfDebugLog( AMAZONSEARCH_DEBUG, "wfRenderAmazonSearch(): can't call api $apiurl.\n" ) );
		return '<!-- amazon>API call failure.</amazon -->';
	}

	/* for debug */	AMAZONSEARCH_DEBUG && ( wfDebugLog( AMAZONSEARCH_DEBUG, "wfRenderAmazonSearch(): API call success.\n$apiurl\n" ) );

	/* decode JSON */
	if( is_null($resapi = json_decode( wfAmazonSearchConvertEncoding( $ret, 'UTF-8'), true ) ) ) {
		/* for debug */	AMAZONSEARCH_DEBUG && ( wfDebugLog( AMAZONSEARCH_DEBUG, "wfRenderAmazonSearch(): can't decode json:\n$json\n" ) );
		return '<!-- amazon>json decode failure.</amazon -->';
	}

	/* no result */
	if( $resapi['SearchResult']['TotalResultCount'] < 1 ) {
		/* for debug */	AMAZONSEARCH_DEBUG && ( wfDebugLog( AMAZONSEARCH_DEBUG, "wfRenderAmazonSearch(): no result.\n" . var_export($resapi, TRUE) . "\n" ) );
		return '<!-- amazon>' . htmlentities($resapi['Error']['Message']) . '</amazon -->';
	}

	/* get result, render start */
	$item = $resapi['SearchResult']['Items'];
	if( isset($item[0]) ) {
		$item = $item[0];
	}
	/* for debug */	AMAZONSEARCH_DEBUG && wfDebugLog( AMAZONSEARCH_DEBUG, "wfRenderAmazonSearch(): get result, first item:\n" . var_export($item, TRUE) . "\n" );

	/* generate embed HTML tags */
	$ret = '<div class="' . AMAZONSEARCH_CLASS_DIV;
	if( !empty($wgAmazonSearchClass['div']) ) {
		$ret .= ' ' . $wgAmazonSearchClass['div'];
	}
	$ret .= '">';

	/* image */
	$ret .= '<a href="' . $item['DetailPageURL'] . '"' . ($wgAmazonSearchTarget ? ' target="' . $wgAmazonSearchTarget . '"' : '') . ' rel="nofollow">';
	$ret .= '<img src="' . $item['Images']['Primary']['Medium']['URL'] . '" width="' . $item['Images']['Primary']['Medium']['Width'] . '" height="' . $item['Images']['Primary']['Medium']['Height'];
	if( !empty($wgAmazonSearchClass['img']) ) {
		$ret .= '" class="' . $wgAmazonSearchClass['img'];
	}
	$ret .= '" /></a>';

	/* item name and link */
	$name = $item['ItemInfo']['Title']['DisplayValue'];
	$maxlen = $wgAmazonSearchMaxLen['name'] ? $wgAmazonSearchMaxLen['name'] : AMAZONSEARCH_NAMEMAXLEN;
	$ret .= '<p';
	if( !empty($wgAmazonSearchClass['name']) ) {
		$ret .= ' class="' . $wgAmazonSearchClass['name'] . '"';
	}
	$ret .= '><a href="' . $item['DetailPageURL'] . '"' . ($wgAmazonSearchTarget ? ' target="' . $wgAmazonSearchTarget . '"' : '') . ' rel="nofollow">';
	$ret .= htmlspecialchars(wfAmazonSearchStrimwidth($name, 0, $maxlen));
	$ret .= '</a></p>';

	/* Author, Artist, Actor or Manufacturer name */
	if( !empty($item['ItemInfo']['ByLineInfo']['Contributors']) ) {
		if( !empty($item['ItemInfo']['ByLineInfo']['Contributors'][1]) ) {
			$text = $item['ItemInfo']['ByLineInfo']['Contributors'][0]['Name'] . ' ' . $item['ItemInfo']['ByLineInfo']['Contributors'][0]['Role'] . ', ' . $item['ItemInfo']['ByLineInfo']['Contributors'][1]['Name'] . ' ' . $item['ItemInfo']['ByLineInfo']['Contributors'][1]['Role'];
		} else {
			$text = $item['ItemInfo']['ByLineInfo']['Contributors']['Name'] . ' ' . $item['ItemInfo']['ByLineInfo']['Contributors']['Role'];
		}
	} elseif(  empty($text = $item['ItemInfo']['ByLineInfo']['Brand']['DisplayValue']) ) {
		$text = ['ItemInfo']['ByLineInfo']['Manufacturer']['DisplayValue'];
	}
	$maxlen = $wgAmazonSearchMaxLen['author'] ? $wgAmazonSearchMaxLen['author'] : AMAZONSEARCH_AUTHORMAXLEN;
	$ret .= '<p';
	if( !empty($wgAmazonSearchClass['author']) ) {
		$ret .= ' class="' . $wgAmazonSearchClass['author'] . '"';
	}
	$ret .= '>';
	$ret .= htmlspecialchars(wfAmazonSearchStrimwidth($text, 0, $maxlen));
	$ret .= '</p>';

	/* link button */
	if( !empty( $caption = wfMessage( 'amazonsearch-sitename-' . $region )->inContentLanguage()->escaped() ) ) {
		$ret .= '<ul';
		if( !empty($wgAmazonSearchClass['buttons']) ) {
			$ret .= ' class="' . $wgAmazonSearchClass['buttons'] . '"';
		}
		$ret .= '><li';
		if( !empty($wgAmazonSearchClass['button']) ) {
			$ret .= ' class="' . $wgAmazonSearchClass['button'] . '"';
		}
		$ret .= '><a href="https://www.amazon.co.jp/gp/search?keywords=' . rawurlencode($value) . '&__mk_ja_JP=%E3%82%AB%E3%82%BF%E3%82%AB%E3%83%8A&tag=' . $atag . '" title="' . wfMessage( 'amazonsearch-button-title', $value, $caption )->inContentLanguage()->escaped() . '"' . ($wgAmazonSearchTarget ? ' target="' . $wgAmazonSearchTarget . '"' : '') . ' rel="nofollow">' . wfMessage( 'amazonsearch-button-caption', $caption )->inContentLanguage()->escaped() . '</a></li></ul>';
	}
	$ret .= "</div>\n";

	/* for debug */	AMAZONSEARCH_DEBUG && ( wfDebugLog( AMAZONSEARCH_DEBUG, "wfRenderYJShopping() complete: \n$ret\n" ) );

	return array( $ret, "markerType" => 'nowiki' );
}


/* mb_convert_encoding() wrapper
 */
function wfAmazonSearchConvertEncoding( $str, $to_encoding )  {
	if( ini_get( 'mbstring.detect_order' ) ) {
		return mb_convert_encoding($str, $to_encoding);
	} else {
		return $str;
	}
}

/* mb_strimwidth() wrapper
 */
function wfAmazonSearchStrimwidth( $str, $start, $width )  {
	if( $width < 0 || strlen($str) <= $width ) {
		return $str;
	} else {
		if( ini_get( 'mbstring.detect_order' ) ) {
			return mb_strimwidth($str, $start, $width, wfMessage( 'amazonsearch-trimmarker' )->inContentLanguage()->plain() );
		} else {
			return substr($str, $start, $width) . wfMessage( 'amazonsearch-trimmarker' )->inContentLanguage()->plain();
		}
	}
}


/* methods for generate Amazon PA-API v5 Signature
 * https://webservices.amazon.co.jp/paapi5/scratchpad/index.html
 */
class AmazonSearchPAAPI5SearchItems {
	private $accessKey = null;
	private $secretKey = null;
	private $host = '';
	private $path = '/paapi5/searchitems';
	private $regionName = null;
	private $serviceName = 'ProductAdvertisingAPI';
	private $httpProtocol = 'https://';
	private $httpMethodName = 'POST';
	private $queryParametes = array();
	private $Headers = array();
	private $payload = '';

	private $HMACAlgorithm = "AWS4-HMAC-SHA256";
	private $aws4Request = "aws4_request";
	private $strSignedHeader = null;
	private $xAmzDate = null;
	private $currentDate = null;

	public function __construct($accessKey, $secretKey, $host, $region) {
		$this->accessKey = $accessKey;
		$this->secretKey = $secretKey;
		$this->xAmzDate = gmdate( "Ymd\THis\Z" );
		$this->currentDate = gmdate( "Ymd" );
		$this->regionName = $region;
		$this->Headers['content-encoding'] = 'amz-1.0';
		$this->Headers['content-type'] = 'application/json; charset=utf-8';
		$this->Headers['host'] = $this->host = $host;
		$this->Headers['x-amz-target'] = 'com.amazon.paapi5.v1.ProductAdvertisingAPIv1.SearchItems';
	}

	public function setPayload($param) {
		$this->payload = json_encode($param, JSON_UNESCAPED_UNICODE);
	}

	public function addHeader($headerName, $headerValue) {
		$this->Headers[$headerName] = $headerValue;
	}

	public function getURL() {
		return $this->httpProtocol . $this->host . $this->path;
	}

	public function createStreamContext() {
		// generate parameters and make sign
		$header = $this->getHeaders();
		// generate http header string
		$headerString = "";
		foreach( $header as $key => $value ) {
			$headerString .= $key . ': ' . $value . "\r\n";
		}
		// generate PHP stream context and return it
		$params = array (
			'http' => array (
				'header' => $headerString,
				'method' => $this->httpMethodName,
				'content' => $this->payload
			)
		);
		/* for debug */	AMAZONSEARCH_DEBUG && ( wfDebugLog( AMAZONSEARCH_DEBUG, "create_stream_context(): out with follows:\n" . var_export($params, TRUE) . "\n" ) );
		return stream_context_create( $params );
	}

	private function prepareCanonicalRequest() {
		$canonicalURL  = $this->httpMethodName . "\n";
		$canonicalURL .= $this->path . "\n" . "\n";
		$signedHeaders = '';
		foreach ( $this->Headers as $key => $value ) {
			$signedHeaders .= $key . ";";
			$canonicalURL .= $key . ":" . $value . "\n";
		}
		$canonicalURL .= "\n";
		$this->strSignedHeader = substr ( $signedHeaders, 0, - 1 );
		$canonicalURL .= $this->strSignedHeader . "\n";
		$canonicalURL .= $this->generateHex ( $this->payload );
		return $canonicalURL;
	}

	private function prepareStringToSign($canonicalURL) {
		$stringToSign = '';
		$stringToSign .= $this->HMACAlgorithm . "\n";
		$stringToSign .= $this->xAmzDate . "\n";
		$stringToSign .= $this->currentDate . "/" . $this->regionName . "/" . $this->serviceName . "/" . $this->aws4Request . "\n";
		$stringToSign .= $this->generateHex ( $canonicalURL );
		return $stringToSign;
	}

	private function calculateSignature($stringToSign) {
		$signatureKey = $this->getSignatureKey( $this->secretKey, $this->currentDate, $this->regionName, $this->serviceName );
		$signature = hash_hmac( "sha256", $stringToSign, $signatureKey, true );
		$strHexSignature = strtolower( bin2hex( $signature ) );
		return $strHexSignature;
	}

	private function getHeaders() {
		$this->Headers['x-amz-date'] = $this->xAmzDate;
		ksort( $this->Headers );

		// Step 1: CREATE A CANONICAL REQUEST
		$canonicalURL = $this->prepareCanonicalRequest();

		// Step 2: CREATE THE STRING TO SIGN
		$stringToSign = $this->prepareStringToSign( $canonicalURL );

		// Step 3: CALCULATE THE SIGNATURE
		$signature = $this->calculateSignature( $stringToSign );

		// Step 4: CALCULATE AUTHORIZATION HEADER
		if ($signature) {
			$this->Headers['Authorization'] = $this->buildAuthorizationString( $signature );
			return $this->Headers;
		}
	}

	private function buildAuthorizationString($strSignature) {
		return $this->HMACAlgorithm . " " . "Credential=" . $this->accessKey . "/" . gmdate( "Ymd" ) . "/" . $this->regionName . "/" . $this->serviceName . "/" . $this->aws4Request . "," . "SignedHeaders=" . $this->strSignedHeader . "," . "Signature=" . $strSignature;
	}

	private function generateHex($data) {
		return strtolower( bin2hex( hash( "sha256", $data, true ) ) );
	}

	private function getSignatureKey($key, $date, $regionName, $serviceName) {
		$kSecret = "AWS4" . $key;
		$kDate = hash_hmac( "sha256", $date, $kSecret, true );
		$kRegion = hash_hmac( "sha256", $regionName, $kDate, true );
		$kService = hash_hmac( "sha256", $serviceName, $kRegion, true );
		$kSigning = hash_hmac( "sha256", $this->aws4Request, $kService, true );
		return $kSigning;
	}
}
?>
